#include "kdialogd.h"
#include <iostream>
#include <kdiroperator.h>
#include <kapplication.h>
#include <qsocketnotifier.h>
#include <kio/netaccess.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <X11/Xlib.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/errno.h>
#include <ksockaddr.h>
#include <kdebug.h>

// from kdebase/kdesu
static int createSocket()
{
    int         sockfd;
    ksocklen_t  addrlen;
    struct stat s;
    const char  *sock=getSockName(1);  // CPD TODO
    int         stat_err=lstat(sock, &s);

    if(!stat_err && S_ISLNK(s.st_mode))
    {
        kdWarning() << "KDialogD - Someone is running a symlink attack on you\n";
        if(unlink(sock))
        {
            kdWarning() << "KDialogD - Could not delete symlink\n";
            return -1;
        }
    }

    if (!access(sock, R_OK|W_OK))
    {
        kdWarning() << "KDialogD - stale socket exists\n";
        if (unlink(sock))
        {
            kdWarning() << "KDialogD - Could not delete stale socket\n";
            return -1;
        }
    }

    sockfd = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sockfd < 0)
    {
        kdError() << "KDialogD - socket(): " << errno << "\n";
        return -1;
    }

    struct sockaddr_un addr;
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, sock, sizeof(addr.sun_path)-1);
    addr.sun_path[sizeof(addr.sun_path)-1] = '\000';
    addrlen = SUN_LEN(&addr);
    if (bind(sockfd, (struct sockaddr *)&addr, addrlen) < 0)
    {
        kdError() << "KDialogD - KDialogD - bind(): " << errno << "\n";
        return -1;
    }

    struct linger lin;
    lin.l_onoff = lin.l_linger = 0;
    if (setsockopt(sockfd, SOL_SOCKET, SO_LINGER, (char *) &lin, sizeof(linger)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_LINGER): " << errno << "\n";
        return -1;
    }
    
    int opt = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char *) &opt, sizeof(opt)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_REUSEADDR): " << errno << "\n";
        return -1;
    }
    opt = 1;
    if (setsockopt(sockfd, SOL_SOCKET, SO_KEEPALIVE, (char *) &opt, sizeof(opt)) < 0)
    {
        kdError() << "KDialogD - setsockopt(SO_KEEPALIVE): " << errno << "\n";
        return -1;
    }

    if(::listen(sockfd, 1)<0)
    {
        kdError() << "KDialogD - listen(1): " << errno << "\n";
        return -1;
    }

    //chmod(sock, 0600);
    return sockfd;
}

static void urls2Local(KURL::List &urls, QStringList &items, QWidget *parent)
{
    KURL::List::Iterator it(urls.begin()),
                         end(urls.end());

    for(; it!=end; ++it)
        if((*it).isLocalFile())
            items.append((*it).path());
        else
        {
            KURL url(KIO::NetAccess::mostLocalURL(*it, parent));

            if(url.isLocalFile())
                items.append(url.path());
            else
                break;
        }
}

KDialogD::KDialogD(QObject *parent)
        : QObject(parent),
          fd(::createSocket())
{
    if(fd<0)
        kdError() << "KDialogD - KDialogD could not create socket\n";
    else
        connect(new QSocketNotifier(fd, QSocketNotifier::Read, this),
                SIGNAL(activated(int)), this, SLOT(newConnection()));
}

KDialogD::~KDialogD()
{
    if(-1!=fd)
        close(fd);
}

void KDialogD::newConnection()
{
    kdDebug() << "KDialogD - New connection\n";

    ksocklen_t         addrlen = 64;
    struct sockaddr_un clientname;
    int                connectedFD;

    if((connectedFD=::accept(fd, (struct sockaddr *) &clientname, &addrlen))>=0)
        connect(new KDialogDClient(connectedFD, this), SIGNAL(error(KDialogDClient *)),
               this, SLOT(deleteConnection(KDialogDClient *)));
}

void KDialogD::deleteConnection(KDialogDClient *client)
{
    kdDebug() << "KDialogD - Delete client\n";
    delete client;
}

KDialogDClient::KDialogDClient(int sock, QObject *parent)
              : QObject(parent),
                fd(sock),
                dlg(NULL),
                accepted(false)
{
    kdDebug() << "KDialogD - new client..." << fd << "\n";
    connect(new QSocketNotifier(fd, QSocketNotifier::Read, this),
            SIGNAL(activated(int)), this, SLOT(read()));
}

KDialogDClient::~KDialogDClient()
{
    kdDebug() << "KDialogD - delete client..." << fd << "\n";
    ::close(fd);
}

void KDialogDClient::close()
{
    emit error(this);
    if(dlg)
    {
        dlg->close();
        dlg->delayedDestruct();
        dlg=NULL;
    }
}

void KDialogDClient::read()
{
    char         request;
    QCString     caption;
    unsigned int xid=0;

    if(!dlg && readData(&request, 1) && request>=(char)OP_FILE_OPEN && request<=(char)OP_FOLDER &&
       readData((char *)&xid, 4) && readString(caption))
    {
        if(OP_FOLDER==(Operation)request)
        {
            QCString intialFolder;

            if(readString(intialFolder))
            {
                initDialog(caption, new KDialogDDirSelectDialog(intialFolder, true, NULL, "folderdialog", false),
                           xid);
                return;
            }
        }
        else
        {
            QCString intialFolder,
                     filter;
            char     overW=0;

            if(readString(intialFolder) && readString(filter) &&
               (OP_FILE_SAVE!=(Operation)request || readData(&overW, 1)))
            {
                initDialog(caption, new KDialogDFileDialog((Operation)request, intialFolder, filter,
                                                           overW ? true : false), xid);
                return;
            }
        }
    }

    kdDebug() << "KDialogD - Comms error, closing connection..." << fd << "\n";
    // If we get here something was wrong, close connection...
    close();
}

void KDialogDClient::finished()
{
    //
    // * finished is emitted when a dialog is ok'ed/cancel'ed/closed
    // * if the user just closes the dialog - neither ok nor cancel are emitted
    // * the dir select dialog doesnt seem to set the QDialog result parameter
    //   when it is accepted - so for this reason if ok is clicked we store an
    //   'accepted' value there, and check for that after the dialog is finished.
    kdDebug() << "KDialogD - finished" << "\n";
    if(dlg && !(accepted || QDialog::Accepted==dlg->result()))
        cancel();
}

void KDialogDClient::ok(const QStringList &items)
{
    int                        num=items.count();
    QStringList::ConstIterator it(items.begin()),
                               end(items.end());
    bool                       error=!writeData((char *)&num, 4);

    for(; !error && it!=end; ++it)
        error=!writeString((*it).local8Bit());

    if(error)
        close();
    else
        accepted=true;
    if(dlg)
        dlg->delayedDestruct();
    dlg=NULL;
}

void KDialogDClient::cancel()
{
    if(dlg)
    {
        int rv=0;

        if(!writeData((char *)&rv, 4))
            close();
        if(dlg)
            dlg->delayedDestruct();
        dlg=NULL;
    }
}

bool KDialogDClient::readData(QCString &buffer, int size)
{
    buffer.resize(size);
    return ::readBlock(fd, buffer.data(), size);
}

bool KDialogDClient::readString(QCString &str)
{
    int size;

    if(!readData((char *)&size, 4))
        return false;

    str.resize(size);
    return readData(str.data(), size);
}

bool KDialogDClient::writeString(const QCString &str)
{
    int size=str.length()+1;

    return writeData((char *)&size, 4) && writeData(str.data(), size);
}

void KDialogDClient::initDialog(const QString &caption, KDialogBase *d, unsigned int xid)
{
    accepted=false;
    dlg=d;

    if(!caption.isEmpty())
        dlg->setPlainCaption(caption);

    if(xid)
    {
        XWindowAttributes attr;
        int               rx, ry;
        Window            junkwin;

        XSetTransientForHint(qt_xdisplay(), dlg->winId(), xid);
        if(XGetWindowAttributes(qt_xdisplay(), xid, &attr))
        {
            XTranslateCoordinates(qt_xdisplay(), xid, attr.root,
                                  -attr.border_width, -16,
                                  &rx, &ry, &junkwin);

            rx=(rx+(attr.width/2))-(dlg->width()/2);
            if(rx<0)
                rx=0;
            ry=(ry+(attr.height/2))-(dlg->height()/2);
            if(ry<0)
                ry=0;
            dlg->move(rx, ry);
        }
    }

    connect(dlg, SIGNAL(ok(const QStringList &)), this, SLOT(ok(const QStringList &)));
    connect(dlg, SIGNAL(finished()), this, SLOT(finished()));
    dlg->show();
}

KDialogDFileDialog::KDialogDFileDialog(Operation op, const QString &startDir, const QString &filter,
                                       bool confirmOw)
                  : KFileDialog(startDir.isEmpty() ? getenv("HOME") : startDir, filter, NULL, NULL, false),
                    itsConfirmOw(confirmOw)
{
    switch(op)
    {
        case OP_FILE_OPEN:
            setOperationMode(KFileDialog::Opening);
            setMode((KFile::Mode)(KFile::File | KFile::ExistingOnly));
            break;
        case OP_FILE_OPEN_MULTIPLE:
            setOperationMode(KFileDialog::Opening);
            setMode((KFile::Mode)(KFile::Files | KFile::ExistingOnly));
            break;
        case OP_FILE_SAVE:
            setOperationMode(KFileDialog::Saving);
            setMode(KFile::File);
            break;
        default:
            break;
    }
}

void KDialogDFileDialog::accept()
{
    kdDebug() << "KDialogD - KDialogDFileDialog::accept" << "\n";
}

void KDialogDFileDialog::slotOk()
{
    setResult(QDialog::Accepted);
    KFileDialog::slotOk();

    kdDebug() << "KDialogD - KDialogDFileDialog::slotOk" << "\n";
    KURL::List  urls;
    QStringList items;
    bool        good=true;

    if(mode()&KFile::Files)
        urls=selectedURLs();
    else
        urls.append(selectedURL());

    if(urls.count())
    {
        urls2Local(urls, items, this);

        if(urls.count()!=items.count())
        {
            KMessageBox::error(this, i18n("You can only access local files."));
            good=false;
        }
        else if(itsConfirmOw && KFileDialog::Saving==operationMode())
            good=!KIO::NetAccess::exists(urls.first(), false, this) ||
                 KMessageBox::Continue==KMessageBox::warningContinueCancel(this,
                                            i18n("File %1 exits.\nDo you want to replace it?")
                                                 .arg(urls.first().prettyURL()),
                                            i18n("File Exists"),
                                            KGuiItem(i18n("Replace"), "filesaveas"), QString::null,
                                            KMessageBox::Notify|KMessageBox::PlainCaption);

        if(good)
        {
            QString filter(currentFilter());

            if(!filter.isEmpty())
                items.append(filter);

            emit ok(items);
            hide();
            KFileDialog::accept();
       }
    }

    if(!good)
        setResult(QDialog::Rejected);
}

KDialogDFileDialog::~KDialogDFileDialog()
{
    kdDebug() << "KDialogD - ~KDialogDFileDialog\n";
}

void KDialogDDirSelectDialog::slotOk()
{
    kdDebug() << "KDialogD - KDialogDDirSelectDialog::slotOk" << "\n";

    KURL::List  urls;
    QStringList items;

    urls.append(url());
    urls2Local(urls, items, this);

    if(urls.count()!=items.count())
        KMessageBox::error(this, i18n("You can only access local folders."));
    else
    {
        emit ok(items);
        hide();
    }
}

KDialogDDirSelectDialog::~KDialogDDirSelectDialog()
{
    kdDebug() << "KDialogD - ~KDialogDDirSelectDialog\n";
}

extern "C"
{
    KDE_EXPORT KDEDModule *create_kdialogd(const QCString &obj)
    {
        return new KDialogDKDED(obj);
    }
};

KDialogDKDED::KDialogDKDED(const QCString &obj)
            : KDEDModule(obj)
{
    new KDialogD(this);
}

#include "kdialogd.moc"

